import manga109api


def test_data_type():
    manga109_root_dir = "tests/data_dummy/"
    p = manga109api.Parser(root_dir=manga109_root_dir)

    for book in p.books:
        annotation = p.get_annotation(book=book)

        # title
        assert isinstance(annotation["title"], str)

        # character
        assert isinstance(annotation["character"], list)
        for character in annotation["character"]:
            assert isinstance(character["@id"], str)
            assert isinstance(character["@name"], str)

        # page
        assert isinstance(annotation["page"], list)
        for page in annotation["page"]:
            assert isinstance(page["@index"], int)
            assert isinstance(page["@width"], int)
            assert isinstance(page["@height"], int)
            
            for obj_type in {"body", "face", "frame", "text"}:
                assert isinstance(page[obj_type], list)
                for obj in page[obj_type]:
                    assert isinstance(obj["@id"], str)
                    assert isinstance(obj["@xmin"], int)
                    assert isinstance(obj["@xmax"], int)
                    assert isinstance(obj["@ymin"], int)
                    assert isinstance(obj["@ymax"], int)

                    if obj_type == "text":
                        assert isinstance(obj["#text"], str)



import pathlib
import json
import xmltodict


class Parser(object):
    def __init__(self, root_dir):
        """
        Manga109 annotation parser

        Args:
            root_dir (str): The path of the root directory of Manga109 data, e.g., 'YOUR_PATH/Manga109_2017_09_28'
        """
        self.root_dir = pathlib.Path(root_dir)
        self.books = []  # book titles

        with (self.root_dir / "books.txt").open("rt", encoding='utf-8') as f:
            self.books = [line.rstrip() for line in f]

    def get_annotation(self, book, annotation_type="annotations"):
        """
        Given a book title, return annotation in the form of dict.

        Args:
            book (str): A title of a book. Should be in self.books.

        Returns:
            annotation (dict): Annotation data consists of dict.
        """
        assert book in self.books
        with (self.root_dir / annotation_type / (book + ".xml")).open("rt", encoding= 'utf-8') as f:
            annotation = xmltodict.parse(f.read())
        annotation = json.loads(json.dumps(annotation))  # OrderedDict -> dict
        annotation = _format_annotation(annotation)
        _convert_str_to_int_recursively(annotation)  # str -> int, for some attributes
        return annotation

    def img_path(self, book, index):
        """
        Given a book title and an index of a page, return the correct image path

        Args:
            book (str): A title of a book. Should be in self.books.
            index (int): An index of a page

        Returns:
            str: A path to the selected image
        """
        assert book in self.books
        assert isinstance(index, int)
        return str((self.root_dir / "images" / book / (str(index).zfill(3) + ".jpg")).resolve())

def _format_annotation(annotation):
    """
    Given annotation data, convert to an easily accessible dict.
    For example, dict['book']['characters']['character'] -> dict['character']

    Args:
        annotation (dict): Annotation data. Root key is 'book'.

    Returns:
        annotation (dict): Annotation data. Root keys are 'title', 'character' and 'page'.
    """

    title = annotation['book']['@title']
    try:
        character = annotation['book']['characters']['character']
    except:
        character = None
    page = annotation['book']['pages']['page']

    if not isinstance(character, list):
        character = [character]
    if not isinstance(page, list):
        page = [page]
    _format_page_dict_style(page)

    return {
        'title': title,
        'character': character,
        'page': page
    }


def _format_page_dict_style(page):
    """
    Format page annotation data. Make page data have the same key, and align the style of dict.
    For example,
      in:  [{'body': [123], 'face': 123, 'frame': []}]
      out: [{'body': [123], 'face': [123], 'frame': [], 'text': []}]

    Args:
        page (dict): Annotation data for all pages including info such as frame, text, etc.
    """
    types = ['body', 'face', 'frame', 'text']
    for i, p in enumerate(page):
        for t in set(types) - set(p.keys()):
            page[i][t] = []
        for t in types:
            if not isinstance(p[t], list):
                page[i][t] = [page[i][t]]


def _convert_str_to_int_recursively(annotation):
    """
    Given annotation data (nested list or dict), convert some attributes from string to integer.
    For example, [{'@xmax': '234', 'id': '0007a8be'}] -> [{'@xmax': 234, 'id': '0007a8be'}]

    Args:
        annotation  (list or dict): Annotation date that consists of list or dict. Can be deeply nested.
    """
    if isinstance(annotation, dict):
        for k, v in annotation.items():
            if k in ['@index', '@width', '@height', '@xmax', '@ymax', '@xmin', '@ymin']:
                annotation[k] = int(v)
            _convert_str_to_int_recursively(v)
    elif isinstance(annotation, list):
        for v in annotation:
            _convert_str_to_int_recursively(v)



import manga109api
from pathlib import Path


def test_img_path():
    manga109_root_dir = "tests/data_dummy/"
    p = manga109api.Parser(root_dir=manga109_root_dir)

    img1 = Path(p.img_path(book="TitleA", index=0)).absolute()
    img2 = Path("tests/data_dummy/images/TitleA/000.jpg").absolute()
    assert(img1 == img2)


def test_format_annotation():
    annotation = {
        'book': {
            '@title': 'AAA',
            'characters': {'character': [
                {'id': '123', 'name': 'yyy'}
            ]},
            'pages': {'page': [
                {'index': 234, 'width': 345, 'height': 456,
                 'frame': [{'id': '567', 'xmin': 11, 'ymin': 22, 'xmax': 33, 'ymax': 44}]}
            ]}
        }
    }
    gt = {
        'title': 'AAA',
        'character': [{'id': '123', 'name': 'yyy'}],
        'page': [
            {'index': 234, 'width': 345, 'height': 456, 'face': [], 'body': [], 'text': [],
             'frame': [{'id': '567', 'xmin': 11, 'ymin': 22, 'xmax': 33, 'ymax': 44}]}
        ]
    }

    ret = manga109api.manga109api._format_annotation(annotation)
    assert(gt == ret)

def test_format_page_dict_style():
    page = [{'body': [123], 'face': 123, 'frame': []}]
    gt = [{'body': [123], 'face': [123], 'frame': [], 'text': []}]
    manga109api.manga109api._format_page_dict_style(page)
    assert(gt == page)

def test_convert_str_to_int_recursively():
    annotation = [{'@xmax': '234', 'id': '0007a8be'}]
    gt = [{'@xmax': 234, 'id': '0007a8be'}]
    manga109api.manga109api._convert_str_to_int_recursively(annotation)
    assert(gt == annotation)
